#
# Copyright (c) 2008-2015 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

__docformat__ = 'restructuredtext'


import json

from pyramid.decorator import reify
from pyramid.events import subscriber
from pyramid.exceptions import NotFound
from pyramid.view import view_config
from transaction.interfaces import ITransactionManager
from z3c.form import field
from z3c.form.interfaces import HIDDEN_MODE, IDataExtractedEvent
from zope.interface import alsoProvides, Invalid, implementer

from pyams_cache.beaker import get_cache
from pyams_form.form import AJAXAddForm, AJAXEditForm, ajax_config
from pyams_form.group import NamedWidgetsGroup
from pyams_pagelet.interfaces import IPagelet, PageletCreatedEvent
from pyams_pagelet.pagelet import pagelet_config
from pyams_portal.interfaces import IPortalContext, IPortalPage, IPortalPortletsConfiguration, \
    IPortalTemplate, IPortalTemplateConfiguration, IPortalTemplateContainer, \
    IPortalTemplateContainerConfiguration, IPortlet, IPortletAddingInfo, IPortletPreviewer, \
    ISlot, ISlotConfiguration, LOCAL_TEMPLATE_NAME, MANAGE_TEMPLATE_PERMISSION
from pyams_portal.portlet import PORTLETS_CACHE_NAME, PORTLETS_CACHE_REGION
from pyams_skin.interfaces import IInnerPage
from pyams_skin.interfaces.viewlet import IMenuHeader, IToolbarAddingMenu
from pyams_skin.layer import IPyAMSLayer
from pyams_skin.viewlet.menu import MenuItem
from pyams_skin.viewlet.toolbar import JsToolbarMenuItem, ToolbarMenuDivider, ToolbarMenuItem
from pyams_template.template import template_config
from pyams_utils.adapter import adapter_config
from pyams_utils.interfaces.data import IObjectData
from pyams_utils.registry import query_utility
from pyams_utils.traversing import get_parent
from pyams_viewlet.manager import viewletmanager_config
from pyams_viewlet.viewlet import viewlet_config
from pyams_zmi.form import AdminDialogAddForm, AdminDialogEditForm
from pyams_zmi.interfaces.menu import IContentManagementMenu, IPropertiesMenu
from pyams_zmi.layer import IAdminLayer
from pyams_zmi.view import AdminView

from pyams_portal import _


@adapter_config(context=(IPortalTemplate, IContentManagementMenu), provides=IMenuHeader)
class PortalTemplateMenuHeader(object):
    """Portal template menu header"""

    def __init__(self, context, menu):
        self.context = context
        self.menu = menu

    @property
    def header(self):
        return _("Template management")


@viewlet_config(name='template-layout.menu',
                context=IPortalTemplate, layer=IAdminLayer,
                manager=IContentManagementMenu, weight=1,
                permission=MANAGE_TEMPLATE_PERMISSION)
@viewletmanager_config(name='template-layout.menu',
                       layer=IAdminLayer,
                       provides=IPropertiesMenu)
@implementer(IPropertiesMenu)
class PortalTemplatePropertiesMenu(MenuItem):
    """Portal template properties menu"""

    label = _("Layout")
    icon_class = 'fa-edit'
    url = '#layout.html'


@pagelet_config(name='layout.html', context=IPortalTemplate, layer=IPyAMSLayer,
                permission=MANAGE_TEMPLATE_PERMISSION)
@template_config(template='templates/layout.pt', layer=IAdminLayer)
@implementer(IInnerPage)
class PortalTemplateLayoutView(AdminView):
    """Portal template main layout configuration view"""

    @property
    def title(self):
        container = get_parent(self.context, IPortalTemplateContainer)
        if container is None:
            context = get_parent(self.context, IPortalContext)
            page = IPortalPage(context)
            if page.use_local_template:
                return _("Local template layout")
            if page.template.name == LOCAL_TEMPLATE_NAME:
                return _("Inherited local template configuration")
            translate = self.request.localizer.translate
            return translate(_("Shared template configuration ({0})")).format(page.template.name)
        return _("Template layout")

    def get_template(self):
        return self.context

    def get_context(self):
        return self.context

    @property
    def can_change(self):
        return self.request.has_permission(MANAGE_TEMPLATE_PERMISSION)

    @reify
    def template_configuration(self):
        return IPortalTemplateConfiguration(self.get_template())

    @reify
    def portlet_configuration(self):
        return IPortalPortletsConfiguration(self.get_context())

    @property
    def selected_portlets(self):
        container = query_utility(IPortalTemplateContainer)
        configuration = IPortalTemplateContainerConfiguration(container)
        utility = self.request.registry.queryUtility
        return filter(lambda x: x is not None,
                      [utility(IPortlet, name=portlet_name)
                       for portlet_name in configuration.toolbar_portlets or ()])

    def get_portlet(self, name):
        return self.request.registry.queryUtility(IPortlet, name=name)

    def get_portlet_add_label(self, portlet):
        translate = self.request.localizer.translate
        return translate(_("Add component: {0}<br />"
                           "Drag and drop button to page template to position new row")).format(
            translate(portlet.label).lower())

    def get_portlet_label(self, name):
        portlet = self.get_portlet(name)
        if portlet is not None:
            return self.request.localizer.translate(portlet.label)
        return self.request.localizer.translate(_("{{ missing portlet }}"))

    def get_portlet_preview(self, portlet_id):
        settings = self.portlet_configuration.get_portlet_configuration(portlet_id).settings
        previewer = self.request.registry.queryMultiAdapter(
            (self.get_context(), self.request, self, settings), IPortletPreviewer)
        if previewer is not None:
            previewer.update()
            return previewer.render()
        return ''


#
# Rows views
#

@viewlet_config(name='add-template-row.menu', context=IPortalTemplate, layer=IAdminLayer,
                view=PortalTemplateLayoutView, manager=IToolbarAddingMenu,
                permission=MANAGE_TEMPLATE_PERMISSION, weight=1)
class PortalTemplateRowAddMenu(JsToolbarMenuItem):
    """Portal template row add menu"""

    label = _("Add row...")
    label_css_class = 'fa fa-fw fa-indent'
    url = 'PyAMS_portal.template.addRow'


@view_config(name='add-template-row.json',
             context=IPortalTemplate, request_type=IPyAMSLayer,
             permission=MANAGE_TEMPLATE_PERMISSION, renderer='json', xhr=True)
def add_template_row(request):
    """Add template raw"""
    config = IPortalTemplateConfiguration(request.context)
    return {'row_id': config.add_row()}


@view_config(name='set-template-row-order.json',
             context=IPortalTemplate, request_type=IPyAMSLayer,
             permission=MANAGE_TEMPLATE_PERMISSION, renderer='json', xhr=True)
def set_template_row_order(request):
    """Set template rows order"""
    config = IPortalTemplateConfiguration(request.context)
    row_ids = map(int, json.loads(request.params.get('rows')))
    config.set_row_order(row_ids)
    return {'status': 'success'}


@view_config(name='delete-template-row.json',
             context=IPortalTemplate, request_type=IPyAMSLayer,
             permission=MANAGE_TEMPLATE_PERMISSION, renderer='json', xhr=True)
def delete_template_row(request):
    """Delete template row"""
    config = IPortalTemplateConfiguration(request.context)
    config.delete_row(int(request.params.get('row_id')))
    return {'status': 'success'}


#
# Slots views
#

@viewlet_config(name='add-template-slot.menu',
                context=IPortalTemplate, layer=IAdminLayer, view=PortalTemplateLayoutView,
                manager=IToolbarAddingMenu, weight=2,
                permission=MANAGE_TEMPLATE_PERMISSION)
class PortalTemplateSlotAddMenu(ToolbarMenuItem):
    """Portal template slot add menu"""

    label = _("Add slot...")
    label_css_class = 'fa fa-fw fa-columns'
    url = 'add-template-slot.html'
    modal_target = True


@pagelet_config(name='add-template-slot.html',
                context=IPortalTemplate, layer=IPyAMSLayer,
                permission=MANAGE_TEMPLATE_PERMISSION)
@ajax_config(name='add-template-slot.json',
             context=IPortalTemplate, layer=IPyAMSLayer,
             permission=MANAGE_TEMPLATE_PERMISSION, base=AJAXAddForm)
class PortalTemplateSlotAddForm(AdminDialogAddForm):
    """Portal template slot add form"""

    @property
    def title(self):
        translate = self.request.localizer.translate
        return translate(_("« {0} »  portal template")).format(self.context.name)

    legend = _("Add slot")
    icon_css_class = 'fa fa-fw fa-columns'

    fields = field.Fields(ISlot)
    edit_permission = None

    def updateWidgets(self, prefix=None):
        super(PortalTemplateSlotAddForm, self).updateWidgets()
        row_id = self.request.params.get('{0}{1}row_id'.format(self.prefix, self.widgets.prefix))
        if row_id:
            self.widgets['row_id'].value = str(int(row_id) + 1)
            self.widgets['row_id'].mode = HIDDEN_MODE

    def createAndAdd(self, data):
        data = data.get(self, data)
        config = IPortalTemplateConfiguration(self.context)
        row_id = data.get('row_id')
        if row_id:
            row_id = row_id - 1
        return config.add_slot(data.get('name'), row_id)

    def get_ajax_output(self, changes):
        return {
            'status': 'callback',
            'callback': 'PyAMS_portal.template.addSlotCallback',
            'options': {
                'row_id': changes[0],
                'slot_name': changes[1]
            }
        }


@subscriber(IDataExtractedEvent, form_selector=PortalTemplateSlotAddForm)
def handle_new_slot_data_extraction(event):
    """Handle new slot form data extraction"""
    config = IPortalTemplateConfiguration(event.form.context)
    data = event.data
    name = data.get('name')
    if name in config.slot_names:
        event.form.widgets.errors += (Invalid(_("Specified name is already used!")),)
    row_id = data.get('row_id')
    if row_id is not None:
        try:
            row_id = int(row_id)
        except ValueError:
            event.form.widgets.errors += (Invalid(_("Row ID must be an integer value!")),)
        else:
            if not 0 < row_id <= config.rows:
                translate = event.form.request.localizer.translate
                event.form.widgets.errors += (Invalid(translate(
                    _("Row ID must be between 1 and {0}!")).format(config.rows)),)


@view_config(name='set-template-slot-order.json',
             context=IPortalTemplate, request_type=IPyAMSLayer,
             permission=MANAGE_TEMPLATE_PERMISSION, renderer='json', xhr=True)
def set_template_slot_order(request):
    """Set template slots order"""
    config = IPortalTemplateConfiguration(request.context)
    order = json.loads(request.params.get('order'))
    for key in order.copy().keys():
        order[int(key)] = order.pop(key)
    config.set_slot_order(order)
    return {'status': 'success'}


@view_config(name='get-slots-width.json',
             context=IPortalTemplate, request_type=IPyAMSLayer,
             permission=MANAGE_TEMPLATE_PERMISSION, renderer='json', xhr=True)
def get_template_slots_width(request):
    """Get template slots width"""
    config = IPortalTemplateConfiguration(request.context)
    return config.get_slots_width(request.params.get('device'))


@view_config(name='set-slot-width.json',
             context=IPortalTemplate, request_type=IPyAMSLayer,
             permission=MANAGE_TEMPLATE_PERMISSION, renderer='json', xhr=True)
def set_template_slot_width(request):
    """Set template slot width"""
    config = IPortalTemplateConfiguration(request.context)
    config.set_slot_width(request.params.get('slot_name'),
                          request.params.get('device'),
                          int(request.params.get('width')))
    return config.get_slots_width(request.params.get('device'))


@pagelet_config(name='slot-properties.html',
                context=IPortalTemplate, layer=IPyAMSLayer,
                permission=MANAGE_TEMPLATE_PERMISSION)
@ajax_config(name='slot-properties.json',
             context=IPortalTemplate, layer=IPyAMSLayer)
class PortalTemplateSlotPropertiesEditForm(AdminDialogEditForm):
    """Slot properties edit form"""

    prefix = 'slot.'

    @property
    def title(self):
        translate = self.request.localizer.translate
        return translate(_("« {0} »  portal template")).format(self.context.name)

    @property
    def legend(self):
        translate = self.request.localizer.translate
        return translate(_("Edit « {0} » slot properties")).format(self.getContent().slot_name)

    fields = field.Fields(ISlotConfiguration).omit('portlet_ids')

    dialog_class = 'modal-large'
    label_css_class = 'control-label col-md-5'
    input_css_class = 'col-md-7'

    edit_permission = MANAGE_TEMPLATE_PERMISSION

    def __init__(self, context, request):
        super(PortalTemplateSlotPropertiesEditForm, self).__init__(context, request)
        self.config = IPortalTemplateConfiguration(context)

    def getContent(self):
        slot_name = self.request.params.get('{0}widgets.slot_name'.format(self.prefix))
        return self.config.slot_config[slot_name]

    def updateWidgets(self, prefix=None):
        super(PortalTemplateSlotPropertiesEditForm, self).updateWidgets(prefix)
        widget = self.widgets.get('slot_name')
        if widget is not None:
            widget.mode = HIDDEN_MODE
        for name in ('prefix', 'suffix'):
            widget = self.widgets.get(name)
            if widget is not None:
                widget.widget_css_class = 'monospace input height-100 text-editor'
                widget.object_data = {
                    'ams-editor-mode': 'ace/mode/html'
                }
                alsoProvides(widget, IObjectData)

    def updateGroups(self):
        self.add_group(NamedWidgetsGroup(self, 'html-codes', self.widgets,
                                         ('prefix', 'suffix'),
                                         legend=_("HTML codes"),
                                         css_class='inner',
                                         switch=True))
        super().updateGroups()

    def get_ajax_output(self, changes):
        if not changes:
            return super(self.__class__, self).get_ajax_output(changes)
        slot_name = self.widgets['slot_name'].value
        slot_config = self.config.slot_config[slot_name]
        return {
            'status': 'success',
            'callback': 'PyAMS_portal.template.editSlotCallback',
            'options': {
                'slot_name': slot_name,
                'width': slot_config.get_width(),
                'css': slot_config.css_class or ''
            }
        }


@view_config(name='delete-template-slot.json',
             context=IPortalTemplate, request_type=IPyAMSLayer,
             permission=MANAGE_TEMPLATE_PERMISSION, renderer='json', xhr=True)
def delete_template_slot(request):
    """Delete template slot"""
    config = IPortalTemplateConfiguration(request.context)
    config.delete_slot(request.params.get('slot_name'))
    return {'status': 'success'}


#
# Portlet views
#

@viewlet_config(name='add-template-portlet.divider',
                context=IPortalTemplate, layer=IAdminLayer, view=PortalTemplateLayoutView,
                manager=IToolbarAddingMenu, weight=10,
                permission=MANAGE_TEMPLATE_PERMISSION)
class PortalTemplateAddMenuDivider(ToolbarMenuDivider):
    """Portal template menu divider"""


@viewlet_config(name='add-template-portlet.menu',
                context=IPortalTemplate, layer=IAdminLayer, view=PortalTemplateLayoutView,
                manager=IToolbarAddingMenu, weight=20,
                permission=MANAGE_TEMPLATE_PERMISSION)
class PortalTemplatePortletAddMenu(ToolbarMenuItem):
    """Portal template portlet add menu"""

    label = _("Add portlet...")
    label_css_class = 'fa fa-fw fa-window-restore'
    url = 'add-template-portlet.html'
    modal_target = True


@pagelet_config(name='add-template-portlet.html',
                context=IPortalTemplate, layer=IPyAMSLayer,
                permission=MANAGE_TEMPLATE_PERMISSION)
@ajax_config(name='add-template-portlet.json',
             context=IPortalTemplate, layer=IPyAMSLayer,
             permission=MANAGE_TEMPLATE_PERMISSION, base=AJAXAddForm)
class PortalTemplatePortletAddForm(AdminDialogAddForm):
    """Portal template portlet add form"""

    @property
    def title(self):
        translate = self.request.localizer.translate
        return translate(_("« {0} »  portal template")).format(self.context.name)

    legend = _("Add portlet")
    icon_css_class = 'fa fa-fw fa-window-restore'

    fields = field.Fields(IPortletAddingInfo)
    edit_permission = None

    def createAndAdd(self, data):
        data = data.get(self, data)
        config = IPortalTemplateConfiguration(self.context)
        return config.add_portlet(data.get('portlet_name'), data.get('slot_name'))

    def get_ajax_output(self, changes):
        configuration = IPortalPortletsConfiguration(self.context)
        settings = configuration.get_portlet_configuration(changes['portlet_id']).settings
        previewer = self.request.registry.queryMultiAdapter(
            (self.context, self.request, self, settings), IPortletPreviewer)
        if previewer is not None:
            previewer.update()
            changes['preview'] = previewer.render()
        return {
            'status': 'callback',
            'callback': 'PyAMS_portal.template.addPortletCallback',
            'options': changes
        }


@view_config(name='drag-template-portlet.json',
             context=IPortalTemplate, request_type=IPyAMSLayer,
             permission=MANAGE_TEMPLATE_PERMISSION, renderer='json', xhr=True)
def drag_template_portlet(request):
    """Drag portlet icon to slot"""
    tmpl_config = IPortalTemplateConfiguration(request.context)
    portlets_config = IPortalPortletsConfiguration(request.context)
    portlet_name = request.params.get('portlet_name')
    slot_name = request.params.get('slot_name')
    changes = tmpl_config.add_portlet(portlet_name, slot_name)
    settings = portlets_config.get_portlet_configuration(changes['portlet_id']).settings
    previewer = request.registry.queryMultiAdapter(
        (request.context, request, request, settings), IPortletPreviewer)
    if previewer is not None:
        previewer.update()
        changes['preview'] = previewer.render()
    return {
        'status': 'callback',
        'close_form': False,
        'callback': 'PyAMS_portal.template.addPortletCallback',
        'options': changes
    }


@view_config(name='set-template-portlet-order.json',
             context=IPortalTemplate, request_type=IPyAMSLayer,
             permission=MANAGE_TEMPLATE_PERMISSION, renderer='json', xhr=True)
def set_template_portlet_order(request):
    """Set template portlet order"""
    order = json.loads(request.params.get('order'))
    order['from'] = int(order['from'])
    order['to']['portlet_ids'] = list(map(int, order['to']['portlet_ids']))
    IPortalTemplateConfiguration(request.context).set_portlet_order(order)
    return {'status': 'success'}


@view_config(name='portlet-properties.html', context=IPortalTemplate, request_type=IPyAMSLayer,
             permission=MANAGE_TEMPLATE_PERMISSION)
class PortalTemplatePortletEditForm(AdminDialogEditForm):
    """Portal template portlet edit form"""

    prefix = 'portlet.'

    dialog_class = 'modal-large'

    def __call__(self):
        request = self.request
        request.registry.notify(PageletCreatedEvent(self))
        portlet_id = int(request.params.get('{0}widgets.portlet_id'.format(self.prefix)))
        portlet_config = IPortalPortletsConfiguration(self.context) \
            .get_portlet_configuration(portlet_id)
        if portlet_config is None:
            raise NotFound()
        editor = self.request.registry.queryMultiAdapter(
            (portlet_config.editor_settings, request), IPagelet, name='properties.html')
        if editor is None:
            raise NotFound()
        request.registry.notify(PageletCreatedEvent(editor))
        editor.ajax_handler = 'portlet-properties.json'
        editor.update()
        return editor()


@view_config(name='portlet-properties.json', context=IPortalTemplate, request_type=IPyAMSLayer,
             permission=MANAGE_TEMPLATE_PERMISSION, renderer='json', xhr=True)
class PortalTemplatePortletAJAXEditForm(AJAXEditForm, PortalTemplatePortletEditForm):
    """Portal template portlet edit form, AJAX renderer"""

    def __call__(self):
        request = self.request
        request.registry.notify(PageletCreatedEvent(self))
        # load portlet config
        portlet_id = int(request.params.get('{0}widgets.portlet_id'.format(self.prefix)))
        portlet_config = IPortalPortletsConfiguration(self.context) \
            .get_portlet_configuration(portlet_id)
        if portlet_config is None:
            raise NotFound()
        # check inheritance
        old_override = portlet_config.inherit_parent
        new_override = request.params.get('{0}widgets.override_parent'.format(self.prefix))
        if new_override:
            portlet_config.inherit_parent = False
        else:
            portlet_config.inherit_parent = True
        # update settings
        changed_override = portlet_config.inherit_parent != old_override
        if changed_override:
            translate = self.request.localizer.translate
            changes = {
                'status': 'success',
                'message': translate(self.successMessage)
            }
        else:
            # get edit output
            editor = request.registry.queryMultiAdapter((portlet_config.editor_settings, request),
                                                        IPagelet, name='properties.json')
            if editor is None:
                raise NotFound()
            changes = editor()
        if changes:
            # clear portlets cache on settings update
            portlets_cache = get_cache(PORTLETS_CACHE_REGION, PORTLETS_CACHE_NAME)
            portlets_cache.clear()
            # we commit before loading previewer to avoid BLOBs "uncommitted changes" error
            ITransactionManager(self.context).commit()
            previewer = request.registry.queryMultiAdapter((self.context, request, self,
                                                            portlet_config.settings),
                                                           IPortletPreviewer)
            if previewer is not None:
                previewer.update()
                changes.update({
                    'callback': 'PyAMS_portal.template.editPortletCallback',
                    'options': {
                        'portlet_id': portlet_id,
                        'inherit_parent': portlet_config.inherit_parent,
                        'preview': previewer.render()
                    }
                })
        if 'autosubmit' in request.params:
            changes['close_form'] = False
        return changes


@view_config(name='delete-template-portlet.json',
             context=IPortalTemplate, request_type=IPyAMSLayer,
             permission=MANAGE_TEMPLATE_PERMISSION, renderer='json', xhr=True)
def delete_template_portlet(request):
    """Delete template portlet"""
    config = IPortalTemplateConfiguration(request.context)
    config.delete_portlet(int(request.params.get('portlet_id')))
    return {'status': 'success'}
